;;; yp-preview.el --- preview function and variable in emacs
;;
;; Copyright (c)
;;
;; This file is for C and C++ code navigation.
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You will not receive a copy of the GNU General Public License with this program,
;; beause I don't provide it. But you can get it 
;; from <http://www.gnu.org/licenses/>.
;;
;; version: 0.2

(defvar pv-buffer-name "*preview*"
  "the buffer name for open to preview define or reference, it will be create in init.")

(defvar pv-show-define-delay-time 0.25
  "when mouse click, delay a time to show the define, when search define need a bit long
time, this can let emacs seems smoothness")

;; variable for search tag, semantic use file name and position 
;; but gtags use file name and line number
(defvar pv-file-name nil
  "the file to be opened")
(defvar pv-position nil
  "the define or reference's position in the file")
(defvar pv-line-number nil
  "the define or reference's line number in the file")

(defvar pv-semantic-or-gtags nil
  "1 means find tag from semantic, 2 means find tag from gtags")
(defvar pv-show-to-preview-or-main-window nil
  "when mouse click, 1 menas show tag in preview window, 2 means show tag in main window")

;; variable for stack
(defvar pv-origin-buffer nil
  "the buffer before jump")
(defvar pv-origin-point nil
  "the position before jump")
(defvar pv-file-stack nil
  "stack for file name, combine with stack for point to find the position to jump back")
(defvar pv-point-stack nil
  "stack for point, combine with stack for file name to find the position to jump back")
(defvar pv-opened-file-stack nil
  "stack record the file opened by preview, if this file is not opened before,
it will be close auto")
(defvar pv-file-exist-stack nil
  " a 't' means the file is opened before, nil means it is never opened")

(defvar pv-preview-buffer-string nil
  "contains the result of 'global -x' if more than one thing is found.")
(defvar pv-preview-buffer-position nil
  "back up the position of preview buffer, when gtags find many result,
we use this variable to jump back")

(define-derived-mode pv-mode
  c-mode "preview")

(define-key pv-mode-map [mouse-1] 'pv-mouse-1-clicked)
(define-key pv-mode-map [mouse-3] 'pv-mouse-3-clikced)

(defun pv-find-in-semantic (dest)
  "search define use semantic tools"
  ;; We have a tag, but in C++, we usually get a prototype instead
  ;; because of header files.  Lets try to find the actual
  ;; implementaion instead.
;;   (when (semantic-tag-prototype-p dest)
;;     (let* ((refs (semantic-analyze-tag-references dest))
;; 	   (impl (semantic-analyze-refs-impl refs t))
;; 	   )
;;       (when impl (setq dest (car impl)))))
  ;; Make sure we have a place to go...
  (setq pv-file-name (semantic-tag-file-name dest))
;;   (if (not (and (or (semantic-tag-with-position-p dest)
;; 		    (semantic-tag-get-attribute dest :line))
;; 		pv-file-name))
;;       (error "Tag %s has no buffer information"
;; 	     (semantic-format-tag-name dest)))

  ;; Once we have the tag, we show it in the preview buffer

  (save-current-buffer
    (set-buffer pv-buffer-name)
    (erase-buffer)
    (insert-file-contents pv-file-name)
    (with-selected-window
        (select-window (car (get-buffer-window-list pv-buffer-name)))
      (setq pv-position (semantic-tag-start dest))
      (goto-char pv-position)))
  (setq pv-semantic-or-gtags 1)  ;; identify we find tag from semantic
  (setq pv-show-to-preview-or-main-window 2)  ;; when mouse click, we show this file in main window
  )

(defun pv-push-mark ()
  "add current buffer name and point to stack, for jump back, and also record the file currently
opened, when jump back, it will be closed if it is only opened by this preview program"
  (setq pv-opened-file-stack (cons pv-file-name pv-opened-file-stack))
  (setq pv-point-stack (cons pv-origin-point  pv-point-stack))
  (setq pv-file-stack (cons (buffer-file-name pv-origin-buffer) pv-file-stack))
  )

(defun pv-show-to-main-window ()
  "open the preview window to main wondow"
  (if (get-file-buffer pv-file-name)
      (setq pv-file-exist-stack (cons t pv-file-exist-stack))
    (setq pv-file-exist-stack (cons nil pv-file-exist-stack)))
  (switch-to-buffer (find-file pv-file-name))
  (cond
   ((= pv-semantic-or-gtags 1)  ;; find tags form semantic
    (goto-char pv-position))
   (t  ;; find tags form gtags
    (goto-line pv-line-number))))

(defun pv-gtags-select-auto ()
  "extract file name and line number from gnu global"
  (save-current-buffer
    (set-buffer pv-buffer-name)
    (goto-char 0)
    (beginning-of-line)
    (if (not (looking-at "[^ \t]+[ \t]+\\([0-9]+\\)[ \t]\\([^ \t]+\\)[ \t]"))
        (error "gtags error")
      ;; gtags-match-string is a function defined in gtags.el
      (setq pv-line-number (string-to-number (gtags-match-string 1)))
      (setq pv-file-name (gtags-match-string 2))))
  (setq pv-file-name (expand-file-name pv-file-name))
  (message "line:%d file:%s" pv-line-number pv-file-name))

(defun pv-gtags-select-from-mouse ()
  "extract file name and line number from gnu global"
  (save-current-buffer
    (set-buffer pv-buffer-name)
    (beginning-of-line)
    (if (not (looking-at "[^ \t]+[ \t]+\\([0-9]+\\)[ \t]\\([^ \t]+\\)[ \t]"))
        (error "gtags error")
      ;; gtags-match-string is a function defined in gtags.el
      (setq pv-line-number (string-to-number (gtags-match-string 1)))
      (setq pv-file-name (gtags-match-string 2))))
  (save-current-buffer
    (set-buffer pv-origin-buffer)
    (setq pv-file-name (expand-file-name pv-file-name)))
  (message "line:%d file:%s" pv-line-number pv-file-name))

;;
;; interactive command
;;

(defun pv-enable ()
  "init preview mode"
  (interactive)
  (get-buffer-create pv-buffer-name)
  (save-current-buffer
    (set-buffer pv-buffer-name)
    (pv-mode)))


(defun pv-find-def-and-delay (point)
  (interactive "d")
  (run-with-idle-timer pv-show-define-delay-time nil 'pv-find-def point))

(defun pv-find-def (point)
  (interactive "d")
  (let* ((ctxt (semantic-analyze-current-context point))
         (pf (and ctxt (reverse (oref ctxt prefix))))
         (current-tag (car pf))
         (be-semantic-tag (semantic-tag-p current-tag))
         (lines nil))
    ;; the above code please reference to semantic-ia-fast-jump function
    (setq pv-origin-point point)
    (setq pv-origin-buffer (current-buffer))
    (cond
     ((and be-semantic-tag
           (semantic-tag-of-class-p current-tag 'variable) 
           (not (semantic-tag-variable-constant-p current-tag)))
      (pv-find-in-semantic current-tag))
     (t
      ;; we try to use global to find tag, if don't find, we try to use semantic again.
      ;; it is mainly useful when we have a function declare but do not have its implement.
      (if be-semantic-tag
          (setq pv-tag-name (semantic-tag-name current-tag))
        (setq pv-tag-name current-tag))
      (save-current-buffer
        (set-buffer pv-buffer-name)
        (erase-buffer))
      ;; when we run "global -x" we can not set preview buffer as the current buffer
      (call-process "global" nil pv-buffer-name nil "-x" pv-tag-name)
      (save-current-buffer
        (set-buffer pv-buffer-name)
        (setq lines (count-lines (point-min) (point-max))))
      ;; lines  indicate how many lines output of "global -x", 0 means gnu global do not find tag, then we
      ;; will call semantic to find again, if there is a function declare, semantic can find it, 1 means
      ;; global find one result, then we shows the contains of the result, others means global find a lot
      ;; of result, then we just keep the global output in preview buffer and let user to select.
      (cond
       ((= 0 lines)
        (if be-semantic-tag
            (pv-find-in-semantic current-tag)))
       ((= 1 lines)
        (pv-gtags-select-auto)
        (save-current-buffer
          (set-buffer pv-buffer-name)
          (erase-buffer)
          (insert-file-contents pv-file-name)
          (with-selected-window
              (select-window (car (get-buffer-window-list pv-buffer-name)))
            (goto-line pv-line-number)))
        (setq pv-semantic-or-gtags 2)  ;; identify we find tag from gtags
        (setq pv-show-to-preview-or-main-window 2))  ;; when mouse click, we show this file in main window
        (t
         (setq pv-semantic-or-gtags 2)
         (setq pv-show-to-preview-or-main-window 1)))))))

(defun pv-find-ref-or-paste (point)
  (interactive "d")
  (if (current-word)
      (pv-find-ref point)
    (yank)))

(defun pv-find-ref (point)
  (interactive "d")
  (setq pv-origin-buffer (current-buffer))
  (setq pv-origin-point point)
  (let* ((tag-name (current-word))
         (lines nil))
    (save-current-buffer
      (set-buffer pv-buffer-name)
      (erase-buffer))
    (call-process "global" nil pv-buffer-name nil "-xr" tag-name)
    (call-process "global" nil pv-buffer-name nil "-xs" tag-name)
    (save-current-buffer
      (set-buffer pv-buffer-name)
      (setq lines (count-lines (point-min) (point-max))))
    (cond
     ((= 0 lines)
      (message "don't find any reference"))
     ((= 1 lines)
      (pv-gtags-select-auto)
      (save-current-buffer
        (set-buffer pv-buffer-name)
        (erase-buffer)
        (insert-file-contents pv-file-name)
        (with-selected-window
            (select-window (car (get-buffer-window-list pv-buffer-name)))
          (goto-line pv-line-number)))
      (setq pv-semantic-or-gtags 2)  ;; identify we find tag from gtags
      (setq pv-show-to-preview-or-main-window 2))  ;; when mous click, we show this file in main window
     (t
      (setq pv-semantic-or-gtags 2)
      (setq pv-show-to-preview-or-main-window 1)))))

(defun pv-pop-mark ()
  "jump back to where we come from"
  (interactive)
  (if (not pv-file-stack) nil
    (let (point file exist)
      (setq exist (car pv-file-exist-stack))
      (setq pv-file-exist-stack (cdr pv-file-exist-stack))
      (if (not exist)
          (progn
            (setq file (car pv-opened-file-stack))
            (setq pv-opened-file-stack (cdr pv-opened-file-stack))
            (kill-buffer (get-file-buffer file))))
      (setq file (car pv-file-stack))
      (setq pv-file-stack (cdr pv-file-stack))
      (setq point (car pv-point-stack))
      (setq pv-point-stack (cdr pv-point-stack))
      (switch-to-buffer (get-file-buffer file))
      (goto-char point)
      )))

(defun pv-mouse-1-clicked ()
  (interactive)
  (cond 
   ((= pv-show-to-preview-or-main-window 1)
    (setq pv-preview-buffer-position (point))
    (setq pv-preview-buffer-string (buffer-string))
    (pv-gtags-select-from-mouse)
    (erase-buffer)
    (insert-file-contents pv-file-name)
    (goto-line pv-line-number)
    (setq pv-semantic-or-gtags 2)  ;; identify we find tag from gtags
    (setq pv-show-to-preview-or-main-window 2))  ;;when mouse click, show it in main window
   ((= pv-show-to-preview-or-main-window 2)
    (pv-show-to-main-window)
    (pv-push-mark))
   (t
    (message "pv nothing to do"))))

(defun pv-mouse-3-clikced ()
  (interactive)
  (erase-buffer)
  (insert pv-preview-buffer-string)
  (goto-char pv-preview-buffer-position)
  (setq pv-show-to-preview-or-main-window 1))

(provide 'yp-preview)

;;; yp-preview.el ends here
